@@ -46,6 +46,32 @@
 //! }
 //! ```
 //!
+//! The following example is how this would be performed with the new declarative inline module syntax
+//! introduced in PyO3 0.23 and above.
+//!
+//! ```rust
+//! # mod test_declarative_example {
+//! use pyo3::prelude::*;
+//! use log::info;
+//!
+//! #[pymodule]
+//! mod my_module{
+//!     use super::*;
+//!
+//!    #[pymodule_init]
+//!     fn init(_m: &Bound<'_, PyModule>) -> PyResult<()> {
+//!         pyo3_log::init();
+//!         Ok(())
+//!     }
+//!
+//!     #[pyfunction]
+//!     fn log_something() {
+//!         info!("Something!");
+//!     }
+//! }
+//! # }
+//! ```
+//!
 //! # Performance, Filtering and Caching
 //!
 //! Ideally, the logging system would always consult the Python loggers to know which messages
@@ -82,7 +108,7 @@
 //! # use pyo3_log::{Caching, Logger};
 //! #
 //! # fn main() -> PyResult<()> {
-//! # Python::with_gil(|py| {
+//! # Python::attach(|py| {
 //! let handle = Logger::new(py, Caching::LoggersAndLevels)?
 //!     .filter(LevelFilter::Trace)
 //!     .filter_target("my_module::verbose_submodule".to_owned(), LevelFilter::Warn)
@@ -145,7 +171,7 @@
 //! fn dont_deadlock(py: Python<'_>) {
 //!     info!("This logs fine");
 //!
-//!     py.allow_threads(|| {
+//!     py.detach(|| {
 //!         let background_thread = thread::spawn(|| {
 //!             info!("This'll not deadlock");
 //!         });
@@ -188,6 +214,7 @@ impl ResetHandle {
 /// What the [`Logger`] can cache.
 #[derive(Copy, Clone, Debug, Eq, PartialEq)]
 #[non_exhaustive]
+#[derive(Default)]
 pub enum Caching {
     /// Disables caching.
     ///
@@ -208,19 +235,14 @@ pub enum Caching {
     /// Therefore, once a `Logger` has been cached, it is possible to decide on the Rust side if a
     /// message would get logged or not. If the message is not to be logged, no Python code is
     /// called and the GIL doesn't have to be acquired.
+    #[default]
     LoggersAndLevels,
 }
 
-impl Default for Caching {
-    fn default() -> Self {
-        Caching::LoggersAndLevels
-    }
-}
-
 #[derive(Debug)]
 struct CacheEntry {
     filter: LevelFilter,
-    logger: PyObject,
+    logger: Py<PyAny>,
 }
 
 impl CacheEntry {
@@ -281,6 +303,9 @@ pub struct Logger {
     /// full paths, with `::` separaters (eg. before converting them from Rust to Python).
     filters: HashMap<String, LevelFilter>,
 
+    /// The prefix to prepend to all log targets
+    prefix: Option<String>,
+
     /// The imported Python `logging` module.
     logging: Py<PyModule>,
 
@@ -307,6 +332,7 @@ impl Logger {
         Ok(Self {
             top_filter: LevelFilter::Debug,
             filters: HashMap::new(),
+            prefix: None,
             logging: logging.into(),
             caching,
             cache: Default::default(),
@@ -379,6 +405,15 @@ impl Logger {
         self
     }
 
+    /// Sets a prefix to prepend to log targets before sending log messages to Python.
+    ///
+    /// This allows for Python-side arrangements where logging configurations are only
+    /// attached to logging names other than the root.
+    pub fn set_prefix(mut self, prefix: &str) -> Self {
+        self.prefix = Some(prefix.replace("::", "."));
+        self
+    }
+
     /// Finds a node in the cache.
     ///
     /// The hierarchy separator is `::`.
@@ -408,10 +443,14 @@ impl Logger {
         py: Python<'_>,
         record: &Record,
         cache: &Option<Arc<CacheNode>>,
-    ) -> PyResult<Option<PyObject>> {
+    ) -> PyResult<Option<Py<PyAny>>> {
         let msg = format!("{}", record.args());
         let log_level = map_level(record.level());
-        let target = record.target().replace("::", ".");
+        let mut target = record.target().replace("::", ".");
+        target = match &self.prefix {
+            Some(prefix) => format!("{}.{}", prefix, target),
+            None => target,
+        };
         let cached_logger = cache
             .as_ref()
             .and_then(|node| node.local.as_ref())
@@ -430,7 +469,42 @@ impl Logger {
         // it. And besides, we can save ourselves few python calls if it's turned off.
         if is_enabled_for(&logger, record.level())? {
             let none = py.None();
-            // TODO: kv pairs, if enabled as a feature?
+
+            #[allow(unused_mut)]
+            let mut extra = py.None().into_bound(py);
+
+            #[cfg(feature = "kv")]
+            if record.key_values().count() > 0 {
+                // write structured data to 'extra', serializing the values
+                use log::kv::{Key, Value, VisitSource};
+                use pyo3::types::{PyDict, PyString};
+
+                struct PyDictVisitor<'p> {
+                    dict: Bound<'p, PyDict>,
+                }
+
+                impl<'kvs, 'p> VisitSource<'kvs> for PyDictVisitor<'p> {
+                    fn visit_pair(
+                        &mut self,
+                        key: Key<'kvs>,
+                        value: Value<'kvs>,
+                    ) -> Result<(), log::kv::Error> {
+                        let py_key = PyString::new(self.dict.py(), key.as_str());
+                        let py_value = PyString::new(self.dict.py(), &value.to_string());
+
+                        let _ = self.dict.set_item(py_key, py_value);
+                        Ok(())
+                    }
+                }
+
+                let mut visitor = PyDictVisitor {
+                    dict: PyDict::new(py),
+                };
+                let _ = record.key_values().visit(&mut visitor);
+
+                extra = visitor.dict.into_any();
+            }
+
             let record = logger.call_method1(
                 "makeRecord",
                 (
@@ -441,8 +515,11 @@ impl Logger {
                     msg,
                     PyTuple::empty(py), // args
                     &none,              // exc_info
+                    &none,              // func
+                    extra,              // extra
                 ),
             )?;
+
             logger.call_method1("handle", (record,))?;
         }
 
@@ -498,7 +575,7 @@ impl Logger {
 
 impl Default for Logger {
     fn default() -> Self {
-        Python::with_gil(|py| {
+        Python::attach(|py| {
             Self::new(py, Caching::LoggersAndLevels).expect("Failed to initialize python logging")
         })
     }
@@ -515,7 +592,7 @@ impl Log for Logger {
         let cache = self.lookup(record.target());
 
         if self.enabled_inner(record.metadata(), &cache) {
-            Python::with_gil(|py| {
+            Python::attach(|py| {
                 // If an exception were triggered before this attempt to log,
                 // store it to the side for now and restore it afterwards.
                 let maybe_existing_exception = PyErr::take(py);
